/**
 * Copyright (C) 2010-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.rackspace;

import static org.jclouds.concurrent.MoreExecutors.sameThreadExecutor;

import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.concurrent.Callable;

import org.apache.log4j.Logger;
import org.dasein.cloud.AbstractCloud;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.ProviderContext;
import org.dasein.cloud.jclouds.rackspace.compute.RackspaceComputeServices;
import org.dasein.cloud.jclouds.rackspace.network.RackspaceNetworkServices;
import org.dasein.cloud.jclouds.rackspace.platform.RackspacePlatformServices;
import org.dasein.cloud.jclouds.rackspace.storage.RackspaceStorageServices;
import org.dasein.util.Retry;
import org.jclouds.blobstore.BlobStoreContext;
import org.jclouds.blobstore.BlobStoreContextFactory;
import org.jclouds.compute.ComputeServiceContextFactory;
import org.jclouds.concurrent.MoreExecutors;
import org.jclouds.concurrent.config.ExecutorServiceModule;
import org.jclouds.loadbalancer.LoadBalancerServiceContextFactory;
import org.jclouds.logging.log4j.config.Log4JLoggingModule;
import org.jclouds.cloudloadbalancers.CloudLoadBalancersAsyncClient;
import org.jclouds.cloudloadbalancers.CloudLoadBalancersClient;
import org.jclouds.cloudservers.CloudServersAsyncClient;
import org.jclouds.cloudservers.CloudServersClient;
import org.jclouds.rest.RestContext;

import com.google.common.collect.ImmutableSet;
import com.google.inject.Module;

public class RackspaceCloud extends AbstractCloud {

	static private final Logger logger = Logger.getLogger(RackspaceCloud.class);
	
	@Override
	public String getCloudName() {
		return "Cloud Servers";
	}
	
    public BlobStoreContext getCloudFilesContext() throws CloudException {
        logger.debug("enter - getCloudFilesContext()");
        try {
            ProviderContext ctx = getContext();
            String endpoint = ctx.getEndpoint();
            String user, key;
            
            try {
                user = new String(ctx.getAccessPublic(), "utf-8");
                key = new String(ctx.getAccessPrivate(), "utf-8");
            }
            catch( UnsupportedEncodingException noway ) {
                logger.error("UTF-8 not supported: " + noway.getMessage());
                return null;
            }
                
            ArrayList<Module> modules = new ArrayList<Module>();
            
            modules.add(new Log4JLoggingModule());
            modules.add(new ExecutorServiceModule(MoreExecutors.sameThreadExecutor(), MoreExecutors.sameThreadExecutor()));
            if( endpoint == null || endpoint.trim().equals("") || endpoint.startsWith("https://auth.api.rackspacecloud.com") ) {
                return (new BlobStoreContextFactory()).createContext("cloudfiles-us", user, key, modules);
            }
            else if( endpoint.startsWith("https://lon.auth.api.rackspacecloud.com") ){
                return (new BlobStoreContextFactory()).createContext("cloudfiles-uk", user, key, modules);                
            }
            else {
                throw new CloudException("Dasein Cloud does not yet support the endpoint: " + endpoint);
            }
        }
        finally {
            logger.debug("exit - getCloudFilesContext()");
        }
    }
    
    public RestContext<CloudServersClient,CloudServersAsyncClient> getCloudServersContext() throws CloudException {
        logger.debug("enter - getCloudServersContext()");
        try {
            return (new Retry<RestContext<CloudServersClient,CloudServersAsyncClient>>()).retry(5, new Callable<RestContext<CloudServersClient,CloudServersAsyncClient>>() {
                @Override
                public RestContext<CloudServersClient,CloudServersAsyncClient> call() throws Exception {
                    ProviderContext ctx = getContext();
                    String endpoint = ctx.getEndpoint();
                    String user, key;
                    
                    try {
                        user = new String(ctx.getAccessPublic(), "utf-8");
                        key = new String(ctx.getAccessPrivate(), "utf-8");
                    }
                    catch( UnsupportedEncodingException noway ) {
                        logger.error("UTF-8 not supported: " + noway.getMessage());
                        return null;
                    }
                    if( endpoint == null || endpoint.trim().equals("") || endpoint.startsWith("https://auth.api.rackspacecloud.com") ) {
                        return new ComputeServiceContextFactory().createContext("cloudservers-us", user, key, ImmutableSet.of(new Log4JLoggingModule(), new ExecutorServiceModule(sameThreadExecutor(), sameThreadExecutor()))).getProviderSpecificContext();
                    }
                    else if( endpoint.startsWith("https://lon.auth.api.rackspacecloud.com") ) {
                        return new ComputeServiceContextFactory().createContext("cloudservers-uk", user, key, ImmutableSet.of(new Log4JLoggingModule(), new ExecutorServiceModule(sameThreadExecutor(), sameThreadExecutor()))).getProviderSpecificContext();                        
                    }
                    else {
                        throw new CloudException("Dasein Cloud does not yet support the endpoint: " + endpoint);                        
                    }
                }
            });
        }
        catch( Exception e ) {
            e.printStackTrace();
            throw new CloudException("Unable to connect to cloud: " + e.getMessage());
        }
        finally {
            logger.debug("exit - getCloudServersContext()");
        }       
    }
    
    public RestContext<CloudLoadBalancersClient,CloudLoadBalancersAsyncClient> getLoadBalancersContext() throws CloudException {
        logger.debug("enter - getCloudServersContext()");
        try {
            return (new Retry<RestContext<CloudLoadBalancersClient,CloudLoadBalancersAsyncClient>>()).retry(5, new Callable<RestContext<CloudLoadBalancersClient,CloudLoadBalancersAsyncClient>>() {
                @Override
                public RestContext<CloudLoadBalancersClient,CloudLoadBalancersAsyncClient> call() throws Exception {
                    ProviderContext ctx = getContext();
                    String endpoint = ctx.getEndpoint();
                    String user, key;
                    
                    try {
                        user = new String(ctx.getAccessPublic(), "utf-8");
                        key = new String(ctx.getAccessPrivate(), "utf-8");
                    }
                    catch( UnsupportedEncodingException noway ) {
                        logger.error("UTF-8 not supported: " + noway.getMessage());
                        return null;
                    }
                    if( endpoint == null || endpoint.trim().equals("") || endpoint.startsWith("https://auth.api.rackspacecloud.com") ) {
                        return new ComputeServiceContextFactory().createContext("cloudload-us", user, key, ImmutableSet.of(new Log4JLoggingModule(), new ExecutorServiceModule(sameThreadExecutor(), sameThreadExecutor()))).getProviderSpecificContext();
                    }
                    throw new CloudException("Dasein Cloud does not yet support the endpoint: " + endpoint);                        
                }
            });
        }
        catch( Exception e ) {
            e.printStackTrace();
            throw new CloudException("Unable to connect to cloud: " + e.getMessage());
        }
        finally {
            logger.debug("exit - getLoadBalancersContext()");
        }       
    }
    
    @Override
    public RackspaceComputeServices getComputeServices() {
        return new RackspaceComputeServices(this);
    }
    
    @Override
    public RackspaceDataCenter getDataCenterServices() {
        return new RackspaceDataCenter(this);
    }
    
    @Override
    public RackspaceNetworkServices getNetworkServices() {
        return new RackspaceNetworkServices(this);
    }
    
    @Override
    public RackspacePlatformServices getPlatformServices() {
        return new RackspacePlatformServices(this);
    }
    
	@Override
	public String getProviderName() {
		return "Rackspace";
	}
	
	@Override
	public RackspaceStorageServices getStorageServices() {
	    return new RackspaceStorageServices(this);
	}
	
    @Override
    public String testContext() {
        try {
            if( !getComputeServices().getVirtualMachineSupport().isSubscribed() ) {
                return null;
            }
            return getContext().getAccountNumber();
        }
        catch( Throwable t ) {
            logger.warn("Failed to test Rackspace connection context: " + t.getMessage());
            if( logger.isDebugEnabled() ) {
                t.printStackTrace();
            }
            return null;
        }
    }
}
