/**
 * Copyright (C) 2010-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.rackspace.compute;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.apache.log4j.Logger;
import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.Tag;
import org.dasein.cloud.compute.Architecture;
import org.dasein.cloud.compute.MachineImage;
import org.dasein.cloud.compute.MachineImageState;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.compute.VirtualMachine;
import org.dasein.cloud.compute.VirtualMachineProduct;
import org.dasein.cloud.compute.VirtualMachineSupport;
import org.dasein.cloud.compute.VmState;
import org.dasein.cloud.compute.VmStatistics;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.jclouds.rackspace.RackspaceCloud;
import org.dasein.util.CalendarWrapper;
import org.dasein.util.Retry;
import org.jclouds.cloudservers.CloudServersAsyncClient;
import org.jclouds.cloudservers.CloudServersClient;
import org.jclouds.cloudservers.domain.Addresses;
import org.jclouds.cloudservers.domain.Flavor;
import org.jclouds.cloudservers.domain.RebootType;
import org.jclouds.cloudservers.domain.ServerStatus;
import org.jclouds.cloudservers.options.CreateServerOptions;
import org.jclouds.cloudservers.options.ListOptions;
import org.jclouds.rest.AuthorizationException;
import org.jclouds.rest.RestContext;

import com.google.common.base.Throwables;

public class RackspaceServer implements VirtualMachineSupport {
	
	static private final Logger logger = Logger.getLogger(RackspaceServer.class);
    
    private RackspaceCloud provider = null;
    
    public RackspaceServer(RackspaceCloud provider) {
        this.provider = provider;
    }

    public void boot(String serverId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("Rackspace does not support the pausing of servers.");
    }

    @Override
    public VirtualMachine clone(String serverId, String intoDcId, String name, String description, boolean powerOn, String ... firewallIds) throws InternalException, CloudException {
        if( !powerOn ) {
            throw new OperationNotSupportedException("Rackspace does not support the pausing of servers.");
        }
        AsynchronousTask<String> task = provider.getComputeServices().getImageSupport().imageVirtualMachine(serverId, name, name);
        VirtualMachine server = getVirtualMachine(serverId);
        String imageId;

        while( !task.isComplete() ) {
            try { Thread.sleep(10000L); }
            catch( InterruptedException e ) { }
        }
        if( task.getTaskError() != null ) {
            throw new CloudException(task.getTaskError());
        }
        imageId = task.getResult();
        
        MachineImage image = provider.getComputeServices().getImageSupport().getMachineImage(imageId);
        long timestamp = System.currentTimeMillis() + (CalendarWrapper.MINUTE * 30L);
        
        while( !image.getCurrentState().equals(MachineImageState.ACTIVE) ) {
            if( System.currentTimeMillis() >= timestamp ) {
                throw new CloudException("Cloning timeout");
            }
            try { Thread.sleep(15000L); }
            catch( InterruptedException e ) { }
            image = provider.getComputeServices().getImageSupport().getMachineImage(imageId);
        }
        return launch(image.getProviderMachineImageId(), server.getProduct(), intoDcId, name, name, null, null, false, false);
    }
    
    private CloudServersClient getClient(RestContext<CloudServersClient,CloudServersAsyncClient> ctx) {
        return ctx.getApi();
    }
    
    private CloudServersAsyncClient getAsyncClient(RestContext<CloudServersClient,CloudServersAsyncClient> ctx) {
        return ctx.getAsyncApi();
    }
    
    @Override
	public String getConsoleOutput(String serverId) throws InternalException, CloudException {
	    return "";
	}

    @Override
	public Iterable<String> listFirewalls(String serverId) throws InternalException, CloudException {
        return new ArrayList<String>();
	}

    @Override
	public String getProviderTermForServer(Locale locale) {
		return "server";
	}

	public VirtualMachine getVirtualMachine(String serverId) throws InternalException, CloudException {
	    for( VirtualMachine server : listVirtualMachines() ) {
	        if( server.getProviderVirtualMachineId().equals(serverId) ) {
	            return server;
	        }
	    }
	    return null;
	}
    
    private VmState getServerState(ServerStatus serverStatus) {
        switch( serverStatus ) {
            case ACTIVE: return VmState.RUNNING;
            case BUILD: return VmState.PENDING;
            case DELETED: return VmState.TERMINATED;
            case ERROR: return VmState.PAUSED;
            case REBOOT: return VmState.REBOOTING;
            case SUSPENDED: return VmState.PAUSED;
            default: return VmState.REBOOTING;
        }
    }
    
	public VmStatistics getVMStatistics(String serverId, long begin, long end) throws InternalException, CloudException {
		return new VmStatistics();
	}
	
	static private long                        lastSizeCacheTime = -1L;
	static private List<VirtualMachineProduct> sizeCache         = null;
	static private VirtualMachineProduct[]     sixtyFours        = null;
	
	VirtualMachineProduct[] getSixtyFours() throws InternalException, CloudException {
        if( sixtyFours == null || (System.currentTimeMillis() >= (lastSizeCacheTime + (CalendarWrapper.MINUTE * 10L))) ) {
            listProducts(Architecture.I64);
        }
        if( sixtyFours == null ) {
            VirtualMachineProduct[] tmp = new VirtualMachineProduct[sizeCache.size()];
            int i = 0;
            
            for( VirtualMachineProduct size : sizeCache ) {
                tmp[i++] = size;
            }
            return tmp;
        }
        return sixtyFours;
	}
	
	@Override
	public Iterable<VirtualMachineProduct> listProducts(Architecture architecture) throws InternalException, CloudException {
        if( sizeCache == null || (System.currentTimeMillis() >= (lastSizeCacheTime + (CalendarWrapper.MINUTE * 10L))) ) {
    	    if( architecture.equals(Architecture.I64) ) {
                RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();
                ArrayList<VirtualMachineProduct> sizes = new ArrayList<VirtualMachineProduct>();
                VirtualMachineProduct[] tmp = null;
                
    	        try {
        	        Collection<Flavor> flavors;
        	        
        	        try {
        	            flavors = getClient(ctx).listFlavors(ListOptions.Builder.withDetails());
        	        }
        	        catch( RuntimeException e ) {
        	            logger.error("Error retrieving flavors of " + architecture + " from Rackspace: " + e.getMessage());
        	            e.printStackTrace();
        	            throw new CloudException(e);
        	        }
        	        tmp = new VirtualMachineProduct[flavors.size()];
        	        int i = 0;
        	        
                    for( Flavor flave : flavors ) {
                        VirtualMachineProduct size = new VirtualMachineProduct();
                        
                        size.setProductId(String.valueOf(flave.getId()));
                        size.setName(flave.getName());
                        size.setDiskSizeInGb(flave.getDisk());
                        size.setDescription(flave.getName());
                        size.setRamInMb(flave.getRam());
                        size.setCpuCount(1);
                        sizes.add(size);
                        tmp[i++] = size;
                    }
    	        }
    	        finally {
    	            ctx.close();
    	        }
    	        lastSizeCacheTime = System.currentTimeMillis();
    	        sizeCache = Collections.unmodifiableList(sizes);
    	        if( tmp != null ) {
    	            sixtyFours = tmp;
    	        }
    	    }
    	    else {
    	        return Collections.emptyList();
    	    }
	    }
	    return sizeCache;
	}
    
    public VirtualMachine launch(String imageId, VirtualMachineProduct size, String dataCenterId, String serverName, String description, String withKey, String inVlanId, boolean withMonitoring, boolean forImaging, String... firewalls) throws InternalException, CloudException {
        return launch(imageId, size, dataCenterId, serverName, description, withKey, inVlanId, withMonitoring, forImaging, firewalls, new Tag[0]);
    }
    
    public VirtualMachine launch(String imageId, VirtualMachineProduct size, String dataCenterId, String serverName, String description, String withKey, String inVlanId, boolean withMonitoring, boolean forImaging, String[] firewalls, Tag ... tags) throws InternalException, CloudException {    
        logger.debug("enter - launch(String, ServerSize, String, String, String, String, boolean, String ...)");
        try {
            Future<org.jclouds.cloudservers.domain.Server> task;
            
            try {
                RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();
                
                try {
                    task = getAsyncClient(ctx).createServer(serverName, Integer.parseInt(imageId), Integer.parseInt(size.getProductId()), new CreateServerOptions[0]);
                }
                catch( Exception e ) {
                    logger.error("Failed to launch " + serverName + ": " + e.getMessage());
                    e.printStackTrace();
                    throw new CloudException("Rackspace server launch failed: " + e.getMessage());
                }
                finally {
                    ctx.close();
                }
            }
            catch( RuntimeException e ) {
                logger.error("Could not create server " + imageId + "/" + size.getProductId() + "/" + serverName + " in Rackspace: " + e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
            if( task == null ) {
                logger.error("No task and no error");
                return null;
            }
            org.jclouds.cloudservers.domain.Server s;
            try {
                s = task.get();
                if( s == null ) {
                    logger.error("No server provided by task");
                    return null;
                }
                return toServer(s);
            }
            catch( InterruptedException e ) {
                e.printStackTrace();
                throw new CloudException("Server launch timed out: " + e.getMessage());
            }
            catch( ExecutionException e ) {
                e.printStackTrace();
                throw new CloudException("Error launching server: " + e.getMessage());
            }
            catch( RuntimeException e ) {
                e.printStackTrace();
                throw new CloudException("Unknown launch exception: " + e.getMessage());
            }
            catch( Error e ) {
                e.printStackTrace();
                throw new CloudException("Unknown launch error: " + e.getMessage());
            }
        }
        finally {
            logger.debug("exit - launch(String, ServerSize, String, String, String, String, boolean, String ...)");
        }
	}

    @Override
	public Iterable<VirtualMachine> listVirtualMachines() throws InternalException, CloudException {
		logger.debug("enter - list()");
		try {
		    ArrayList<VirtualMachine> servers = new ArrayList<VirtualMachine>();
		    Collection<org.jclouds.cloudservers.domain.Server> rawServers;
		    

            final RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();

            try {
                Retry<Collection<org.jclouds.cloudservers.domain.Server>> retry = new Retry<Collection<org.jclouds.cloudservers.domain.Server>>();

                rawServers = retry.retry(5, new Callable<Collection<org.jclouds.cloudservers.domain.Server>>() {
                    @Override
                    public Collection<org.jclouds.cloudservers.domain.Server> call() throws Exception {
                        return getClient(ctx).listServers(ListOptions.Builder.withDetails());
                    }
                    
                });
            }
            catch( Exception e ) {
                Throwable t = Throwables.getRootCause(e);
                
                e.printStackTrace();
                if( t != null ) {
                    throw new CloudException(t);
                }
                throw new CloudException(e);                    
            }
            finally {
                ctx.close();
            }

		    for( org.jclouds.cloudservers.domain.Server server : rawServers ) { 
		        VirtualMachine s = toServer(server);
		        
		        if( s != null ) {
		            servers.add(s);
		        }
		    }
		    return servers;
		}
		finally {
		    logger.debug("exit - list()");
		}
	}
    
	@Override
	public String[] mapServiceAction(ServiceAction action) {
		return new String[0];
	}

    @Override
	public void enableAnalytics(String serverId) throws InternalException, CloudException {
	    // NO-OP
	}

    @Override
    public void pause(String serverId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("Rackspace does not support the pausing of servers.");
    }

    @Override
    public void reboot(String serverId) throws CloudException, InternalException {
        logger.debug("enter - reboot(String)");
        try {
            try {
                RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();

                try {
                    getClient(ctx).rebootServer(Integer.parseInt(serverId), RebootType.SOFT);
                }
                finally {
                    ctx.close();
                }
            }
            catch( RuntimeException e ) {
                logger.error("Rackspace failed to reboot the serve " + serverId + ": " + e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            logger.debug("exit - reboot(String)");
        }
    }
    
    @Override
	public void terminate(String serverId) throws InternalException, CloudException {
	    logger.debug("enter - stop(String)");
	    try {
	        VirtualMachine vm = getVirtualMachine(serverId);
	        int timeout = 0;
	        while (vm == null && timeout < 10) {
	        	timeout++;
	        	try { Thread.sleep(15000L); }
	            catch( InterruptedException e ) { }
	            vm = getVirtualMachine(serverId);
	        }
	        if (vm == null) {
	        	throw new CloudException("Virtual machine disappeared.");
	        }
	        while( vm.getCurrentState().equals(VmState.PENDING) ) {
	            try { Thread.sleep(15000L); }
	            catch( InterruptedException e ) { }
	            vm = getVirtualMachine(serverId);
	        }
	        try {
                RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();

                try {
                    getClient(ctx).deleteServer(Integer.parseInt(serverId));
                }
                finally {
                    ctx.close();
                }
	        }
	        catch( RuntimeException e ) {
	            logger.error("Rackspaced failed to terminate the server " + serverId + ": " + e.getMessage());
	            e.printStackTrace();
	            throw new CloudException(e);
	        }
	    }
	    finally {
	        logger.debug("exit - stop(String)");
	    }
	}
	
	private VirtualMachine toServer(org.jclouds.cloudservers.domain.Server rackspaceServer) throws InternalException, CloudException {
	    VirtualMachine server = new VirtualMachine();
	    String str;
	    
	    server.setProviderOwnerId(provider.getContext().getAccountNumber());
	    server.setPersistent(true);
	    server.setArchitecture(Architecture.I64);
	    server.setCreationTimestamp(0L);
        server.setProviderRegionId(provider.getContext().getRegionId());
	    if( provider.getContext().getRegionId().equals("main") ) {
	        server.setProviderDataCenterId("1");
	    }
	    else {
	        server.setProviderDataCenterId("lon1");
	    }
	    server.setLastBootTimestamp(0L);
	    server.setLastPauseTimestamp(-1L);
	    server.setTerminationTimestamp(-1L);
	    server.setClonable(true);
	    server.setImagable(true);
	    server.setPausable(false);
	    server.setPersistent(true);
	    str = rackspaceServer.getAdminPass();
	    if( str != null ) {
	        server.setRootPassword(str);
	    }
        server.setName(rackspaceServer.getName());
        server.setProduct(getProduct(String.valueOf(rackspaceServer.getFlavorId())));
	    server.setProviderVirtualMachineId(String.valueOf(rackspaceServer.getId()));
	    server.setProviderMachineImageId(String.valueOf(rackspaceServer.getImageId()));
	    server.setCurrentState(getServerState(rackspaceServer.getStatus()));
	    server.setDescription(rackspaceServer.getMetadata().get("Server Label"));
	    if( server.getDescription() == null ) {
	        server.setDescription(rackspaceServer.getHostId());
	        if( server.getDescription() == null ) {
	            server.setDescription(server.getName());
	        }
	    }
	    Addresses addresses = rackspaceServer.getAddresses();
	    if( addresses != null ) {
            Set<String> list = addresses.getPrivateAddresses();
            String[] addrs = new String[list.size()];
            int i =0;
            
            for( String addr : list ) {
                addrs[i++] = addr;
            }
            server.setPrivateIpAddresses(addrs);
            
            list = addresses.getPublicAddresses();
            addrs = new String[list.size()];
            i = 0;
            for( String addr : list ) {
                addrs[i++] = addr;
            }
            server.setPublicIpAddresses(addrs);
	    }
	    server.setPlatform(Platform.guess(server.getName()));
	    if( server.getPlatform().equals(Platform.UNKNOWN) ) {
	        server.setPlatform(Platform.guess(server.getDescription()));
	        if( server.getPlatform().equals(Platform.UNKNOWN) ) {
	            server.setPlatform(Platform.UNIX);
	        }
	    }
	    return server;
	}
	
	public void unmonitor(String serverId) throws InternalException, CloudException {
	    // NO-OP
	}

    @Override
    public Iterable<VmStatistics> getVMStatisticsForPeriod(String serverId, long start, long end) throws InternalException, CloudException {
        return new ArrayList<VmStatistics>();
    }

    @Override
    public void disableAnalytics(String vmId) throws InternalException, CloudException {
        // NO-OP
    }

    @Override
    public VirtualMachineProduct getProduct(String productId) throws InternalException, CloudException {
        for( VirtualMachineProduct product : listProducts(Architecture.I64) ) {
            if( product.getProductId().equals(productId) ) {
                return product;
            }
        }
        for( VirtualMachineProduct product : listProducts(Architecture.I32) ) {
            if( product.getProductId().equals(productId) ) {
                return product;
            }
        }    
        return null;
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();

        try {
            getClient(ctx).listServers(ListOptions.Builder.withDetails());
            return true;
        }
        catch( AuthorizationException e ) {
            return false;
        }
        catch( RuntimeException e ) {
            e.printStackTrace();
            throw new InternalException(e);
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public boolean supportsAnalytics() throws CloudException, InternalException {
        return false;
    }  
}
